﻿using System;
using System.IO;
using System.Security.Cryptography;
using System.Text;

namespace E4GuiTool
{
	internal class FTool
	{
		private const string XorSecret = "Any probable_VAL!";
		private const int XorSecretSize = 17;
		private const int PadSize = 16;

		public static ulong Decrypt(Stream inStream, Stream outStream)
		{
			// read tail
			var ubTail = new byte[16];
			inStream.Seek(-ubTail.Length, SeekOrigin.End);
			inStream.Read(ubTail, 0, ubTail.Length);

			// check tail (could be done in a more nice way)
			if (ubTail[0] != 'C' || ubTail[1] != 'F' || ubTail[2] != 'C' || ubTail[3] != 0x00)
				throw new FormatException("Tail part 1 invalid.");

			if (ubTail[4] != 0x03 || ubTail[5] != 0x00 || ubTail[6] != 0x10 || ubTail[7] != 0x00)
				throw new FormatException("Tail part 2 invalid.");

			if (ubTail[12] != 0x02 || ubTail[13] != 0x00 || ubTail[14] != 0x00 || ubTail[15] != 0x00)
				throw new FormatException("Tail part 3 invalid.");

			// get size of plain config
			var size = (uint) (ubTail[8] + (ubTail[9] << 8) + (ubTail[10] << 16) + (ubTail[11] << 24));

			// XOR stuff
			var ubXor = new byte[XorSecretSize];
			var ubXorRound = new byte[XorSecretSize];
			byte[] xorSecretBytes = Encoding.ASCII.GetBytes(XorSecret);
			Buffer.BlockCopy(xorSecretBytes, 0, ubXor, 0, ubXor.Length);

			// known start of plain config
			byte[] ubPlain = Encoding.ASCII.GetBytes("#\x0D\x0A# Gen");

			// back to start of file, read first 8 bytes to determine SUID
			var ubSuid = new byte[8];

			inStream.Seek(0, SeekOrigin.Begin);

			inStream.Read(ubSuid, 0, ubSuid.Length);

			// we know the expected result (ubPlain) and can so determine SUID
			for (int i = 0; i < ubSuid.Length; ++i)
				ubSuid[i] ^= ubPlain[i];

			var ubOffs = (byte) ((ubSuid[0]/2) - ubXor[0]);

			for (int i = 0; i < ubSuid.Length; ++i)
				ubSuid[i] = (byte) (ubSuid[i] - (ubXor[0] + ubOffs) - ubXor[i]);

			ulong suid = BitConverter.ToUInt64(ubSuid, 0);

			// back to start of file
			inStream.Seek(0, SeekOrigin.Begin);

			// build XOR bytes
			for (int i = 0; i < ubXor.Length; ++i)
				ubXor[i] += ubSuid[i & 7];

			// read file (in single byte steps)
			var ub = new byte[1];
			int idx = 0;
			int j;

			// XOR modification round
			Buffer.BlockCopy(ubXor, 0, ubXorRound, 0, ubXorRound.Length);
			for (j = 0; j < ubXorRound.Length; j++)
				ubXorRound[j] += ubXor[idx];
			idx++;

			uint cnt = 0;
			int t = 0;
			while (inStream.Read(ub, 0, ub.Length) == ub.Length)
			{
				cnt++;

				// check if all plain config bytes have been read
				if (cnt > size)
					break;

				// "decrypt"
				ub[0] ^= ubXorRound[t];

				// write result
				outStream.Write(ub, 0, ub.Length);

				t++;
				if (t >= ubXorRound.Length)
				{
					t = 0;
					// XOR modification round
					Buffer.BlockCopy(ubXor, 0, ubXorRound, 0, ubXorRound.Length);
					for (j = 0; j < ubXorRound.Length; j++)
						ubXorRound[j] += ubXor[idx];

					idx++;
					if (idx >= ubXor.Length)
						idx = 0;
				}
			}

			// TODO: should we check the MD5 hash as an additional step ?
			return suid;
		}


		public static void Encrypt(Stream inStream, Stream outStream, ulong suid)
		{
			byte[] suidBytes = BitConverter.GetBytes(suid);
			var ubXor = new byte[XorSecretSize];
			var ubXorRound = new byte[XorSecretSize];

			// build XOR bytes
			byte[] xorSecretBytes = Encoding.ASCII.GetBytes(XorSecret);
			Buffer.BlockCopy(xorSecretBytes, 0, ubXor, 0, ubXor.Length);

			for (int i = 0; i < ubXor.Length; i++)
				ubXor[i] += suidBytes[i & 7];

			// read file (in single byte steps)
			var ub = new byte[1];
			int idx = 0;
			int j;

			// XOR modification round
			Buffer.BlockCopy(ubXor, 0, ubXorRound, 0, ubXorRound.Length);
			for (j = 0; j < ubXorRound.Length; j++)
				ubXorRound[j] += ubXor[idx];
			idx++;

			// MD5 hash
			MD5 md5 = MD5.Create();
			md5.Initialize();
			// four "secret" start bytes for hash
			ub[0] = 0x05;
			md5.TransformBlock(ub, 0, ub.Length, null, 0);
			ub[0] = 0x9A;
			md5.TransformBlock(ub, 0, ub.Length, null, 0);
			ub[0] = 0x63;
			md5.TransformBlock(ub, 0, ub.Length, null, 0);
			ub[0] = 0x17;
			md5.TransformBlock(ub, 0, ub.Length, null, 0);

			uint cnt = 0; // number of bytes in plain config
			int t = 0;
			while (inStream.Read(ub, 0, ub.Length) == ub.Length)
			{
				cnt++; // cnt plain config bytes

				// hash of plain config
				md5.TransformBlock(ub, 0, ub.Length, null, 0);

				// "encrypt"      
				ub[0] ^= ubXorRound[t];

				// write result
				outStream.Write(ub, 0, ub.Length);

				t++;
				if (t >= ubXorRound.Length)
				{
					t = 0;

					// XOR modification round
					Buffer.BlockCopy(ubXor, 0, ubXorRound, 0, ubXorRound.Length);
					for (j = 0; j < ubXorRound.Length; j++)
						ubXorRound[j] += ubXor[idx];

					idx++;
					if (idx >= ubXor.Length)
						idx = 0;
				}
			}

			md5.TransformFinalBlock(new byte[0], 0, 0);

			// pad to 16 byte block size
			// probably we could pad with any value but its done using
			// the XOR keystream like in the original config files


			// XOR modification round
			t = 0;
			Buffer.BlockCopy(ubXor, 0, ubXorRound, 0, ubXorRound.Length);

			for (j = 0; j < ubXorRound.Length; j++)
				ubXorRound[j] += ubXor[idx];

			for (t = 0; t < PadSize - ((int) cnt%PadSize); t++)
			{
				ub[0] = ubXorRound[t];
				// write result
				outStream.Write(ub, 0, ub.Length);
			}

			// write 16 byte MD5 hash
			byte[] md5Hash = md5.Hash;
			outStream.Write(md5Hash, 0, md5Hash.Length);

			// build and write trailer, most of it is fixed
			var trailer = new byte[16];

			trailer[0] = (byte) 'C';
			trailer[1] = (byte) 'F';
			trailer[2] = (byte) 'C';
			trailer[3] = 0x00;

			trailer[4] = 0x03;
			trailer[5] = 0x00;
			trailer[6] = 0x10;
			trailer[7] = 0x00;

			// original plain config size

			trailer[8] = (byte) (cnt & 0xFF);
			trailer[9] = (byte) ((cnt >> 8) & 0xFF);
			trailer[10] = (byte) ((cnt >> 16) & 0xFF);
			trailer[11] = (byte) ((cnt >> 24) & 0xFF);

			trailer[12] = 0x02;
			trailer[13] = 0x00;
			trailer[14] = 0x00;
			trailer[15] = 0x00;

			outStream.Write(trailer, 0, trailer.Length);
		}
	}
}