#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "md5.h"

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef void* PVOID;
typedef int BOOL;

#define TRUE    1
#define FALSE   0

// used to generate the XOR stream

#define XOR_SECRET  "Any probable_VAL!"
const int XOR_SECRET_SIZE = 17;

BOOL decrypt(FILE *fIn, FILE *fOut)
{   
    // read tail
    
    BYTE ubTail[16];
    
    if(fseek(fIn, -(int)sizeof(ubTail), SEEK_END) != 0)
    {
        printf("fseek() error\n");
        return FALSE;
    }
    
    if(fread(ubTail, 1, sizeof(ubTail), fIn) != sizeof(ubTail))
    {
        printf("fread() error\n");
        return FALSE;
    }
    
    // check tail (could be done in a more nice way)
    
    if(ubTail[0] != 'C' || ubTail[1] != 'F' || ubTail[2] != 'C' ||ubTail[3] != 0x00)
    {
        printf("Tail part 1 invalid\n");
        return FALSE;
    }
   
    if(ubTail[4] != 0x03 || ubTail[5] != 0x00 || ubTail[6] != 0x10 || ubTail[7] != 0x00)
    {
        printf("Tail part 2 invalid\n");
        return FALSE;
    }
      
    if(ubTail[12] != 0x02 || ubTail[13] != 0x00 || ubTail[14] != 0x00 || ubTail[15] != 0x00)
    {
        printf("Tail part 3 invalid\n");
        return FALSE;
    }
    
    // get size of plain config
    
    DWORD size = ubTail[8] + (ubTail[9] << 8) + (ubTail[10] << 16) + (ubTail[11] << 24);

    // XOR stuff
            
    BYTE ubXor[XOR_SECRET_SIZE];    
    BYTE ubXorRound[XOR_SECRET_SIZE];
    memcpy(ubXor, XOR_SECRET, sizeof(ubXor));
    
    // known start of plain config

    BYTE ubPlain[8];    
    memcpy(ubPlain, "#\x0D\x0A# Gen", sizeof(ubPlain));
    
    // back to start of file, read first 8 bytes to determine SUID

    BYTE ubSuid[8];
    
    if(fseek(fIn, 0, SEEK_SET) != 0)
    {
        printf("fseek() error\n");
        return FALSE;
    }
    
    if(fread(ubSuid, 1, sizeof(ubSuid), fIn) != sizeof(ubSuid))
    {
        printf("fread() error\n");
        return FALSE;
    }
    
    // we know the expected result (ubPlain) and can so determine SUID
    
    int i;
    for(i = 0; i < sizeof(ubSuid); i++)
    {
        ubSuid[i] ^= ubPlain[i];
    }
    
    BYTE ubOffs = (ubSuid[0] / 2) - ubXor[0];
    
    for(i = 0; i < sizeof(ubSuid); i++)
    {
        ubSuid[i] = (ubSuid[i] - (ubXor[0] + ubOffs) - ubXor[i]);
    }
    
    DWORD dw1 = (ubSuid[0] << 24) + (ubSuid[1] << 16) + (ubSuid[2] << 8) + ubSuid[3];
    DWORD dw2 = (ubSuid[4] << 24) + (ubSuid[5] << 16) + (ubSuid[6] << 8) + ubSuid[7];

    printf("SUID: 0x%08X 0x%08X", dw1, dw2);
        
    // back to start of file
    
    if(fseek(fIn, 0, SEEK_SET) != 0)
    {
        printf("fseek() error\n");
        return FALSE;
    }

    // build XOR bytes
    
    for(i = 0; i < sizeof(ubXor); i++)
        ubXor[i] += ubSuid[i & 7];
      
    // read file (in single byte steps)
      
    BYTE ub[1];
    int idx = 0;
    int j;
         
    // XOR modification round

    memcpy(ubXorRound, ubXor, sizeof(ubXorRound));
    for(j = 0; j < sizeof(ubXorRound); j++)
         ubXorRound[j] += ubXor[idx];
    idx++;
   
    DWORD cnt = 0;
    i = 0;
    while(fread(ub, 1, sizeof(ub), fIn) == sizeof(ub))
    {
        cnt++;

        // check if all plain config bytes have been read
                
        if(cnt > size)
            break;

        // "decrypt"

        ub[0] ^= ubXorRound[i];
              
        // write result

        if(fwrite(ub, 1, sizeof(ub), fOut) != sizeof(ub))
        {
            printf("fwrite() error\n");
            return FALSE;
        }
       
        i++;       
        if(i >= sizeof(ubXorRound))
        {
            i = 0;
          
            // XOR modification round

            memcpy(ubXorRound, ubXor, sizeof(ubXorRound));
            for(j = 0; j < sizeof(ubXorRound); j++)
                 ubXorRound[j] += ubXor[idx];

            idx++;
            if(idx >= sizeof(ubXor))
                idx = 0;
        }
    }

    // TODO: should we check the MD5 hash as an additional step ?

    return TRUE;
}

BOOL encrypt(FILE *fIn, FILE *fOut, BYTE *ubSuid)
{
    BYTE ubXor[XOR_SECRET_SIZE];
    BYTE ubXorRound[XOR_SECRET_SIZE];

    // build XOR bytes
    
    memcpy(ubXor, XOR_SECRET, sizeof(ubXor));
    
    int i;
    for(i = 0; i < sizeof(ubXor); i++)
        ubXor[i] += ubSuid[i & 7];
      
    // read file (in single byte steps)
      
    BYTE ub[1];
    int idx = 0;
    int j;
         
    // XOR modification round

    memcpy(ubXorRound, ubXor, sizeof(ubXorRound));
    for(j = 0; j < sizeof(ubXorRound); j++)
         ubXorRound[j] += ubXor[idx];
    idx++;
   
    // MD5 hash
   
    MD5_CTX mdContext;
   
    MD5Init(&mdContext);
   
    // four "secret" start bytes for hash

    ub[0] = 0x05;
    MD5Update(&mdContext, ub, sizeof(ub));
    ub[0] = 0x9A;
    MD5Update(&mdContext, ub, sizeof(ub));
    ub[0] = 0x63;
    MD5Update(&mdContext, ub, sizeof(ub));
    ub[0] = 0x17;
    MD5Update(&mdContext, ub, sizeof(ub));
   
    DWORD cnt = 0; // number of bytes in plain config
    i = 0;
    while(fread(ub, 1, sizeof(ub), fIn) == sizeof(ub))
    {
        cnt++; // cnt plain config bytes
       
        // hash of plain config

        MD5Update(&mdContext, ub, sizeof(ub));

        // "encrypt"       

        ub[0] ^= ubXorRound[i];

        // write result
              
        if(fwrite(ub, 1, sizeof(ub), fOut) != sizeof(ub))
        {
            printf("fwrite() error\n");
            return FALSE;
        }
       
        i++;       
        if(i >= sizeof(ubXorRound))
        {
            i = 0;
          
            // XOR modification round

            memcpy(ubXorRound, ubXor, sizeof(ubXorRound));
            for(j = 0; j < sizeof(ubXorRound); j++)
                 ubXorRound[j] += ubXor[idx];

            idx++;
            if(idx >= sizeof(ubXor))
                idx = 0;
        }
    }

    MD5Final(&mdContext);
   
    // pad to 16 byte block size
    // probably we could pad with any value but its done using
    // the XOR keystream like in the original config files
   
    const int PAD_SIZE = 16;

    // XOR modification round

    i = 0;            
    memcpy(ubXorRound, ubXor, sizeof(ubXorRound));
    for(j = 0; j < sizeof(ubXorRound); j++)
        ubXorRound[j] += ubXor[idx];
            
    for(i = 0; i < PAD_SIZE - ((int)cnt % PAD_SIZE); i++)
    {
        ub[0] = ubXorRound[i];

        // write result
       
        if(fwrite(ub, 1, sizeof(ub), fOut) != sizeof(ub))
        {
            printf("fwrite() error\n");
            return FALSE;
        }
    }

    // write 16 byte MD5 hash

    if(fwrite(mdContext.digest, 1, sizeof(mdContext.digest), fOut) != sizeof(mdContext.digest))
    {
        printf("fwrite() error\n");
        return FALSE;
    }
   
    // build and write trailer, most of it is fixed
   
    BYTE trailer[16];
   
    trailer[0] = 'C';
    trailer[1] = 'F';
    trailer[2] = 'C';
    trailer[3] = 0x00;
   
    trailer[4] = 0x03;
    trailer[5] = 0x00;
    trailer[6] = 0x10;
    trailer[7] = 0x00;
   
    // original plain config size

    trailer[ 8] = (BYTE)(cnt & 0xFF);
    trailer[ 9] = (BYTE)((cnt >> 8) & 0xFF);
    trailer[10] = (BYTE)((cnt >> 16) & 0xFF);
    trailer[11] = (BYTE)((cnt >> 24) & 0xFF);
   
    trailer[12] = 0x02;
    trailer[13] = 0x00;
    trailer[14] = 0x00;
    trailer[15] = 0x00;
   
    if(fwrite(trailer, 1, sizeof(trailer), fOut) != sizeof(trailer))
    {
        printf("fwrite() error\n");
        return FALSE;
    }
    
    return TRUE;
}

void usage()
{
    printf("usage: ftool -d <file in> <file out>\n");
    printf("       ftool -e <file in> <file out> suid1 suid2\n");
}

int main(int argc, char **argv)
{
    BOOL bDecrypt;
    
    if(argc == 4 && strcmp("-d", argv[1]) == 0)
    {
        bDecrypt = TRUE;
    }
    else if(argc == 6 && strcmp("-e", argv[1]) == 0)
    {
        bDecrypt = FALSE;
    }
    else
    {
        usage();
        return 1;
    }
   
    FILE *fIn = fopen(argv[2], "rb");
    if(fIn == NULL)
    {
        printf("Cannot open input file %s\n", argv[2]);
        return 1;
    }
   
    FILE *fOut = fopen(argv[3], "wb");
    if(fOut == NULL)
    {
        printf("Cannot open output file %s\n", argv[3]);
        fclose(fIn);
        return 1;
    }
    
    if(bDecrypt)
        decrypt(fIn, fOut);
    else
    {
        BYTE ubSuid[8];

        // get 8 byte SUID from two DWORD parameters    
        // a 64-bit type would match better
    
        DWORD dw1 = strtoul(argv[4], NULL, 0);
        DWORD dw2 = strtoul(argv[5], NULL, 0);
        
        ubSuid[0] = (BYTE)(dw1 >> 24);
        ubSuid[1] = (BYTE)(dw1 >> 16);
        ubSuid[2] = (BYTE)(dw1 >>  8);
        ubSuid[3] = (BYTE)(dw1);
        
        ubSuid[4] = (BYTE)(dw2 >> 24);
        ubSuid[5] = (BYTE)(dw2 >> 16);
        ubSuid[6] = (BYTE)(dw2 >>  8);
        ubSuid[7] = (BYTE)(dw2);
        
        encrypt(fIn, fOut, ubSuid);
    }
         
    fclose(fIn);
    fclose(fOut);

    return 0; 
}